/*-
 * Copyright (c) 2010 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/param.h>

#include <dirent.h>
#include <inttypes.h>
#include <netpgp.h>
#include <regex.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <threshold.h>
#include <time.h>
#include <unistd.h>

#include "otk.h"

/* copy from one file to another */
static int
copy(const char *fromdir, const char *from, const char *todir, const char *to)
{
	struct stat	 st;
	size_t		 size;
	char		 fromname[MAXPATHLEN];
	char		 toname[MAXPATHLEN];
	char		*mem;
	FILE		*fpr;
	FILE		*fpw;
	int		 wc;
	int		 cc;

	(void) snprintf(fromname, sizeof(fromname), "%s/%s", fromdir, from);
	(void) snprintf(toname, sizeof(toname), "%s/%s", todir, to);
	if ((fpr = fopen(fromname, "r")) == NULL) {
		(void) fprintf(stderr, "can't copy from '%s'\n", fromname);
		return 0;
	}
	if ((fpw = fopen(toname, "w")) == NULL) {
		(void) fprintf(stderr, "can't copy to '%s'\n", toname);
		(void) fclose(fpr);
		return 0;
	}
	(void) fstat(fileno(fpr), &st);
	size = (size_t)st.st_size;
	mem = mmap(NULL, size, PROT_READ, MAP_PRIVATE | MAP_FILE, fileno(fpr), 0);
	if (mem == MAP_FAILED) {
		(void) fprintf(stderr, "mmap failed '%s'\n", fromname);
		(void) fclose(fpr);
		(void) fclose(fpw);
		return 0;
	}
	for (cc = 0 ; (wc = write(fileno(fpw), &mem[cc], size - cc)) > 0 ; cc += wc) {
	}
	(void) fclose(fpw);
	(void) fclose(fpr);
	(void) munmap(mem, size);
	return 1;
}

/* symlink function (with stdarg) */
static int
vsymlink(const char *contents, const char *fmt, ...)
{
	va_list	args;
	char	name[MAXPATHLEN];

	va_start(args, fmt);
	vsnprintf(name, sizeof(name), fmt, args);
	va_end(args);
	return symlink(contents, name);
}

/* read the contents of a symbolic link */
static int
vreadlink(char *contents, size_t size, const char *fmt, ...)
{
	va_list	args;
	char	name[MAXPATHLEN];
	int	cc;

	va_start(args, fmt);
	vsnprintf(name, sizeof(name), fmt, args);
	va_end(args);
	cc = readlink(name, contents, size);
	contents[cc] = 0x0;
	return cc;
}

/* varargs stat function */
static int
vstat(struct stat *st, const char *op, const char *fmt, ...)
{
	va_list	args;
	char	name[MAXPATHLEN];

	va_start(args, fmt);
	vsnprintf(name, sizeof(name), fmt, args);
	va_end(args);
	return (strcmp(op, "stat") == 0) ? stat(name, st) : lstat(name, st);
}

/* chmod a file */
static int
vsyscall(mode_t mode, const char *op, const char *fmt, ...)
{
	va_list	args;
	char	name[MAXPATHLEN];

	va_start(args, fmt);
	vsnprintf(name, sizeof(name), fmt, args);
	va_end(args);
	return (strcmp(op, "chmod") == 0) ? chmod(name, mode) == 0 : mkdir(name, mode) == 0;
}

/* real basename(3) which strips suffix too */
static char *
realbasename(const char *name, const char *suffix, char *out, size_t size)
{
	const char	*last;
	const char	*s;
	int	 	 sufflen;

	if ((s = strrchr(name, '/')) == NULL) {
		s = name;
	} else {
		s += 1;
	}
	sufflen = strlen(suffix);
	last = &s[strlen(s)];
	if (strcmp(suffix, last - sufflen) == 0) {
		last -= sufflen;
	}
	(void) snprintf(out, size, "%.*s", (int)(last - s), s);
	return out;
}

/* remove 'd' and all files/dirs under it */
static int
recursive_rm(const char *d)
{
	struct dirent	*dp;
	struct stat	 st;
	char		 name[MAXPATHLEN];
	DIR		*dirp;
	int		 ok;

	if ((dirp = opendir(d)) == NULL) {
		(void) fprintf(stderr, "can't open directory '%s'\n", d);
		return 0;
	}
	for (ok = 1 ; (dp = readdir(dirp)) != NULL ; ) {
		if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0) {
			continue;
		}
		(void) snprintf(name, sizeof(name), "%s/%s", d, dp->d_name);
		if (lstat(name, &st) < 0) {
			(void) fprintf(stderr, "can't lstat '%s'\n", name);
			continue;
		}
		switch(st.st_mode & S_IFMT) {
		case S_IFREG:
		case S_IFLNK:
			if (unlink(name) < 0) {
				(void) fprintf(stderr, "can't unlink '%s'\n", name);
				ok = 0;
			}
			break;
		case S_IFDIR:
			if (!recursive_rm(name)) {
				ok = 0;
			}
			break;
		default:
			break;
		}
	}
	(void) closedir(dirp);
	if (ok) {
		ok = (rmdir(d) == 0);
	}
	return ok;
}

/* write a share to a file */
static int
writeshare(const char *share, size_t size, const char *fmt, ...)
{
	va_list	 args;
	size_t	 cc;
	size_t	 wc;
	char	 f[MAXPATHLEN];
	FILE	*fp;

	va_start(args, fmt);
	(void) vsnprintf(f, sizeof(f), fmt, args);
	va_end(args);
	if ((fp = fopen(f, "w")) == NULL) {
		(void) fprintf(stderr, "writeshare: can't write to '%s'\n", f);
		return 0;
	}
	for (cc = 0 ; cc < size && (wc = write(fileno(fp), &share[cc], size - cc)) > 0 ; cc += wc) {
	}
	(void) fclose(fp);
	return 1;
}

/**************************************************************************/
/* exported functions from here */
/**************************************************************************/

/* share the file 'f' */
int
otk_share(otk_t *otk, const char *prog, const char *f)
{
	threshold_t	 threshold;
	struct stat	 st;
	netpgp_t	 netpgp;
	unsigned	 i;
	size_t		 cc;
	time_t		 t;
	char		 keyrings[MAXPATHLEN];
	char		 secret[MAXPATHLEN];
	char		 sharedir[MAXPATHLEN];
	char		 scratch[MAXPATHLEN];
	char		 d[MAXPATHLEN];
	char		 desc[1024];
	char		 cmd[8192];
	char		*share;

	(void) memset(&netpgp, 0x0, sizeof(netpgp));
	(void) memset(&threshold, 0x0, sizeof(threshold));
	if (otk->threshold == 0 || otk->shares == 0) {
		(void) fprintf(stderr, "%s: threshold and shares must be non-zero\n", prog);
		return 0;
	}
	t = time(NULL);
	(void) snprintf(desc, sizeof(desc), "%s shared %u/%u by %s %.24s",
		(otk->anon) ? "[secret]" : f, otk->threshold, otk->shares,
		(otk->anon) ? "[user]" : getenv("USER"), ctime(&t));
	(void) fprintf(stderr, "Generating a one-time key\n");
	if (otk->splitdir == NULL) {
		(void) snprintf(d, sizeof(d), "/tmp/share.XXXXXX");
		otk->splitdir = mkdtemp(d);
	}
	if (stat(otk->splitdir, &st) < 0) {
		mkdir(otk->splitdir, 01700);
	}
	/* generate the one-time key */
	netpgp_set_homedir(&netpgp, otk->splitdir, NULL, 1);
	netpgp_setvar(&netpgp, "userid checks", "skip");
	netpgp_init(&netpgp);
	netpgp_generate_key(&netpgp, desc, (int)otk->numbits);
	/* get generated userid */
	(void) snprintf(otk->keyid, sizeof(otk->keyid), "%s",
		netpgp_getvar(&netpgp, "generated userid"));
	/* encrypt the secret */
	(void) snprintf(keyrings, sizeof(keyrings), "%s/%s", otk->splitdir, otk->keyid);
	netpgp_end(&netpgp);
	(void) memset(&netpgp, 0x0, sizeof(netpgp));
	netpgp_set_homedir(&netpgp, keyrings, NULL, 1);
	netpgp_init(&netpgp);
	vsyscall(0644, "chmod", "%s/pubring.gpg", keyrings);
	(void) snprintf(secret, sizeof(secret), "%s/secret.gpg", keyrings);
	netpgp_encrypt_file(&netpgp, otk->keyid, f, secret, 0);
	/* share the private key */
	thresh_init(&threshold, "ssss", (uint8_t)otk->threshold, (uint8_t)otk->shares);
	(void) snprintf(scratch, sizeof(scratch), "%s/secring.gpg", keyrings);
	thresh_split(&threshold);
	copy(keyrings, "pubring.gpg", otk->splitdir, "pubring.gpg");
	for (i = 0 ; i < otk->shares ; i++) {
		(void) snprintf(sharedir, sizeof(sharedir), "%s/share%.3d", otk->splitdir, i);
		mkdir(sharedir, 01700);
		vsymlink(otk->keyid, "%s/keyid", sharedir);
		copy("", secret, sharedir, "secret.gpg");
		copy(keyrings, "pubring.gpg", sharedir, "pubring.gpg");
		thresh_get_share(&threshold, i, (void **)(void *)&share, &cc);
		writeshare(share, cc, "%s/secring.gpg.share", sharedir);
		(void) snprintf(cmd, sizeof(cmd),
			"tar -C %s -czf %s/share%.3d.tar.gz share%.3d",
			otk->splitdir, otk->splitdir, i, i);
		system(cmd); /* XXX */
		vsyscall(0600, "chmod", "%s/share%.3d.tar.gz", otk->splitdir, i);
		recursive_rm(sharedir);
	}
	thresh_end(&threshold);
	(void) fprintf(stderr, "Shared secrets are in: %s\n", otk->splitdir);
	recursive_rm(keyrings);
	return 1;
}

/* attempt to recover the secret */
int
otk_recover(otk_t *otk, const char *prog, int argc, char **argv)
{
	threshold_t	 threshold;
	struct stat	 st;
	netpgp_t	 netpgp;
	size_t		 size;
	char		*recoverdir;
	char		 cmd[8192];
	char		 both[MAXPATHLEN];
	char		 keyring[MAXPATHLEN];
	char		 scratch[MAXPATHLEN];
	char		 d[MAXPATHLEN];
	char		*share;
	int		 i;

	(void) snprintf(d, sizeof(d), "/tmp/recover.XXXXXX");
	recoverdir = mkdtemp(d);
	(void) memset(otk->keyid, 0x0, sizeof(otk->keyid));
	(void) memset(&threshold, 0x0, sizeof(threshold));
	if (optind == argc) {
		(void) fprintf(stderr, "%s: no files to recover\n", prog);
		return 0;
	}
	for (i = optind ; i < argc ; i++) {
		(void) snprintf(cmd, sizeof(cmd), "tar -C %s -xzf %s", recoverdir, argv[i]);
		system(cmd); /* XXX */
		(void) realbasename(argv[i], ".tar.gz", scratch, sizeof(scratch));
		if (otk->keyid[0] == 0x0) {
			vreadlink(otk->keyid, sizeof(otk->keyid), "%s/%s/keyid", recoverdir, scratch);
		}
		(void) snprintf(keyring, sizeof(keyring), "%s/%s", recoverdir, otk->keyid);
		if (vstat(&st, "stat", "%s", keyring) < 0) {
			vsyscall(01700, "mkdir", "%s", keyring);
		}
		(void) snprintf(both, sizeof(both), "%s/%s", recoverdir, scratch);
		if (vstat(&st, "stat", "%s/secret.gpg", keyring) < 0) {
			copy(both, "secret.gpg", keyring, "secret.gpg");
		}
		if (vstat(&st, "stat", "%s/pubring.gpg", keyring) < 0) {
			copy(both, "pubring.gpg", keyring, "pubring.gpg");
		}
		(void) snprintf(scratch, sizeof(scratch), "%s/secring.gpg.share", both);
		thresh_add_share(&threshold, i - optind, scratch, -1);
	}
	thresh_combine(&threshold);
	thresh_get_share(&threshold, THRESH_JOIN_DEST, (void **)(void *)&share, &size);
	writeshare(share, size, "%s/secring.gpg", keyring);
	(void) memset(&netpgp, 0x0, sizeof(netpgp));
	netpgp_set_homedir(&netpgp, keyring, NULL, 1);
	netpgp_setvar(&netpgp, "userid", otk->keyid);
	netpgp_init(&netpgp);
	(void) snprintf(scratch, sizeof(scratch), "%s/secret.gpg", keyring);
	netpgp_decrypt_file(&netpgp, scratch, otk->out, 0);
	netpgp_end(&netpgp);
	return 1;
}
