/* 
 * secsplit.c
 *
 * Shamir secret sharing
 *
 * Based on "How to Share a Secret", by Adi Shamir, Communications of
 * the ACM, November, 1979, Volume 22, Number 11, page 612.
 *
 * Original copyright:
 * Copyright (C) 1993 Hal Finney, 74076.1041@compuserve.com.
 * This software is being placed in the public domain.
 *
 * This program divides a file into n pieces, such that any k of them
 * are sufficient to reconstruct the original file, but that k-1
 * pieces give NO information about the original file (except its
 * length).
 *
 * It has been written for and tested on DOS and Unix systems.
 *
 * To split up a file, use the command; secsplit n k file where n is
 * the number of pieces to split it up to, and k is the number of
 * pieces needed to reconstruct it. k should be <= n and > 1.  If you
 * get n and k backwards the program will swap them for you so don't
 * worry too much about remembering the order.
 *
 * The program will output to file.001, file.002, .... If the file has
 * an extension (e.g. "file.c") the extension will be stripped off
 * before the ".001", etc., are added (so "file.c" will also output to
 * "file.001", etc.).
 *
 * To reconstruct a file, use the command:
 *    secsplit k file.*
 * or
 *    secsplit k file file1 file2...
 *
 * k should be >= the k used when the file was split; the minimum
 * number of pieces needed to reconstruct the file.  If you have too
 * few pieces then the program won't give an error, but you'll get the
 * wrong answer.
 *
 * The first command form is for DOS or other systems which won't
 * expand the ".*" for you; the program scans for file.000, file.001,
 * etc., and uses the first k of them that it finds.  In the second
 * form, the number of files given should be at least k, and again the
 * first k of them will be used.
 *
 * The output in the first form will be file.out; in the second form
 * it will be the first file on the command line, stripped of its
 * extension, and with ".out" added, so generally it will be file.out
 * too.
 *
 * Shamir's algorithm relies on cryptographically strong, unguessable,
 * random numbers.  This version of the program uses the /dev/urandom
 * interface to the kernel random number pool present on the free
 * Unices.
 *
 *
 * Revision history:
 *    Version 1.0	October 23, 1993
 *
 *    Version 1.1	October 24, 1993
 *       Added IDEA-based random-number-generator, initialized by MD5
 *       of input file, plus the time of day.
 *
 *    Version 1.2	March 31, 2000 Damien Miller <djm@mindrot.org>
 *       Removed IDEA-based random-number generator in favour of 
 *       /dev/random. Removed MSDOSisms. Rearranged source code and 
 *       formatting according to OpenBSD style(9)
 *
 *
 * The file formats used for output are as follows.  Each file starts
 * with one byte which is the index, from 1 through n, of that file.
 * This is the x value used for the polynomial evaluation in Shamir's
 * algorithm.  The files then consist of a series of 16-bit values
 * (high byte first), which are the result of applying Shamir's
 * splitting algorithm to the input file taken in 16-bit chunks.  The
 * prime used is slightly less than 2^16, meaning that input data
 * values close to 2^16 get turned into a pair of values (see below
 * for more information on this expansion).
 *
 * If the input file is of even length, the output files will each by
 * of that length+1 (because of the 1 byte at the beginning).  If the
 * input file is of odd length, the input is padded with a random byte
 * and processed normally to get a pair of output bytes, then each
 * output file is padded with an extra random byte to indicate this
 * fact.  So output files which have an odd length correspond to input
 * files with an even length, and vice versa.
 *
 * The output files could have encoded k and n information, but this
 * could be helpful to an attacker (he would know when he was close to
 * having enough pieces to reconstruct the file if he knew k).  So the
 * user is required to remember this himself.
 *
 * The file formats are system-independent so files split on one kind
 * of machine should be able to be reassembled on another kind of
 * machine.
 *
 * One complication exists with this code.  The algorithm must work
 * mod a prime which is bigger than all the data.  We want the data to
 * be about 16 bits therefore, so that we can multiply two numbers and
 * have them fit in a 32-bit int.  This means that the prime has to be
 * less than 2^16 in order for them to fit.  The largest such prime is
 * 65521.  This means that any data value > 65521 can't be processed
 * directly; instead, I split it into two values, a 65520, followed by
 * value-65520.  Also, 65520 itself is split as 65520 followed by 0.
 * This complication is pretty well limited to the input routine on
 * splitting and the output routine on assembly.
 *
 * The purpose of the "magic" is to xor the incoming value with a
 * different number each time.  This is purely intended to deal with
 * the problem that values of d >= limit-1 must be stored as four
 * bytes rather than two.  I was worried that some file might have a
 * lot of 0xff's which would balloon in size.  So I wanted to xor with
 * a rather random number.  Then, I change the random number each time
 * in order to make it less likely that some other file would happen
 * to have a lot of ballooning values.  This way it would be very rare
 * that a file managed to track my magic number in such a way that
 * many d's were over the size threshold.
 *
 * (The "magic" is not intended to increase the security or the
 * cryptographic strength of the algorithm in any way; it is purely to
 * keep the size of the output files from being much bigger than the
 * input.)
 *
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/mman.h>

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "threshold.h"
#include "internal.h"

#define PRIME 65521

#define S4_KMAX 48
#define S4_NMAX 48

#define S4_IMAGIC 0x8a31
#define S4_DMAGIC 0x1347

#define S4_SMALLEST_INPUT	5

/* Multiplicative inverses of 1-48 mod PRIME */
static const int invtab[] = {
	    1, 32761, 43681, 49141, 52417, 54601, 56161, 57331,
	58241, 58969, 11913, 60061, 60481, 60841, 61153, 61426,
	42396, 61881,  6897, 62245, 62401, 38717, 11395, 62791,
	49796, 63001, 41254, 63181, 58743, 63337, 25363, 30713,
	 3971, 21198, 63649, 63701, 54896, 36209, 63841, 63883,
	43148, 63961,  6095, 52119, 64065, 38458, 43216, 64156
};

typedef struct s4_t {
	unsigned	  extra;	/* extra from last call */
	unsigned	  magic;	/* magic number */
	int		  have_extra;	/* 16-bit entity input */
	int		  oddflag;	/* last return included a pad */
	int		  seen[S4_KMAX]; /* if seen, then not the first */
	unsigned	  next_d1[S4_KMAX]; /* for two byte lookahead */
	uint8_t		  sharesc;	/* # of shares */
} s4_t;

/******************* Code related to splitting *********************/

/* Return a random number from 0 to n-1. */
#define CRANDOM(x)	(uint32_t)(random() % (x))

/*
 * Evaluate the given polynomial, n coefficients, at point x=i.
 * Do it mod the specified modulus.
 *
 * poly = Polynomial coefficients
 * n = # coefficients (order of polynomial + 1)
 * i = Point to evaluate it at
 * mod = Modulus for evaluation
 */
static uint32_t
eval(uint32_t *poly, unsigned n, unsigned i, uint32_t mod)
{
	u_int32_t prod;		/* Accumulated product */
	int j;					/* index */

	prod = poly[n - 1];
	for (j = (n - 1); --j >= 0; ) {
		prod *= i;
		prod += poly[j];
		prod %= mod;
	}
	return prod;
}

/*
 * Return a 16-bit value from file f_in, but limit it to be less than limit.
 * Anything >= limit-1 gets returned as two consecutive values (on 2 calls).
 * Return -1 on EOF, or -2 if the previous return value had been padded
 * because the file had an odd # bytes.
 */
static uint32_t
get_limited_16(s4_t *s4, thresh_str_t *str, uint32_t limit)
{
	uint32_t	d;
	int		c1;
	int		c2;

	/* First check for leftover from last time */
	if (s4->have_extra) {
		s4->have_extra = 0;
		return s4->extra;
	}
	/* Check if last return included a pad */
	if (s4->oddflag) {
		return (uint32_t)-2;
	}
	/* Read data (bigendian), do the magic */
	if (str->c == str->io.size) {
		return (uint32_t)-1;
	}
	c1 = str->io.base[str->c++];
	if (str->c == str->io.size) {
		c2 = CRANDOM(0x100);
		s4->oddflag = 1;
	} else {
		c2 = str->io.base[str->c++];
	}
	d = ((c1 & 0xff) << 8) + (c2 & 0xff);
	d ^= s4->magic;
	s4->magic = (s4->magic + S4_DMAGIC) & 0xffff;
	/*
	 * If over the limit, return limit-1 as a code for that, and remember
	 * to return the rest next time.
	 */
	if (d >= (limit - 1)) {
		s4->have_extra = 1;
		s4->extra = d - (limit - 1);
		d = limit - 1;
	}
	return d;
}

/*
 * Given a 16-bit value d, less than mod, split it into sharesc files such
 * that any k of them can reconstruct it.
 */
static void
split_out(uint32_t d, unsigned sharesc, unsigned k, uint32_t mod, thresh_str_t *shares)
{
	uint32_t	poly[S4_KMAX];
	uint32_t	di;
	unsigned	i;
	unsigned	j;

	poly[0] = d;
	for (j = 1; j < k; j++) {
		poly[j] = CRANDOM(mod);
	}
	for (i = 0; i < sharesc; i++) {
		di = eval(poly, k, i + 1, mod);
		shares[i].io.base[shares[i].c++] = (di >> 8) & 0xff;
		shares[i].io.base[shares[i].c++] = (di & 0xff);
	}
}

#define S4_MAGIC	"s4"

/*
 * Split the specified input file into sharesc output files, such that
 * any k of them are sufficient to reconstruct the input.  mod is the
 * largest prime < 2^16.  This is the main routine for the splitting case.
 *
 * f_in = Input file handle
 * f_out[] = Output file handles
 * sharesc = Number of output files
 * k = Threshhold for re-assembl
 * mod = Modulus for calculations
 */
static void
split(s4_t *s4, thresh_str_t *str, unsigned sharesc, unsigned k, uint32_t mod, thresh_str_t *shares)
{
	thresh_head_t	head;
	unsigned	i;
	int		d;

	thresh_fmt_header(&head, S4_MAGIC, k, sharesc, str->io.size);
	for (i = 0; i < sharesc; i++) {
		/* Prefix each file with "x" coordinate, 1 byte */
		head.coeff = i + 1;
		/* add header */
		(void) memcpy(&shares[i].io.base[shares[i].c], &head, sizeof(head));
		shares[i].c = sizeof(head);
	}
	for (;;) {
		if ((d = get_limited_16(s4, str, PRIME)) == -1) {
			break;
		}
		if (d == -2) {
			/* Odd flag - pad output files with a random byte to remember */
			for (i = 0; i < sharesc; i++) {
				shares[i].io.base[shares[i].c++] = CRANDOM(0x100);
			}
			break;
		}
		split_out((unsigned)d, sharesc, k, mod, shares);
	}
}

/******************* Code related to assembly *********************/

/*
 * Return the multiplicative inverse of small positive or negative value
 * x modulo the current prime.  Do a table lookup for speed. 
 *
 * x = Small value to find inverse of
 */
static uint32_t
inverse(int x)
{
	int neg = 0;

	if (x < 0) {
		x = -x;
		neg = 1;
	}
	if (x < 1 || x > (int)(sizeof(invtab)/sizeof(invtab[0]))) {
		fprintf(stderr, "inverse out of range: %d\n", x);
		exit(EXIT_FAILURE);
	}
	return (neg) ? PRIME - invtab[x-1] : invtab[x-1];
}

/*
 ( Interpolate the polynomial specified at x and y coordinates
 * in the array of size n, at x=i.  Do it mod the specified modulus.
 * This algorithm is from Knuth, The Art of Computer Programming, Vol. 2,
 * Seminumerical Algorithms, section 4.6.4, Evaluation of Polynomials,
 * equation 43 and 44, page 485 (bottom) in 1981 hardcover edition.
 * Knuth has his indices go from 0 to n; mine go from 0 to n-1, a slight
 * notational change.
 *
 * i = x coord of interpolated point
 * x[] = x coordinates of known points
 * y[] = y coordinates of known points
 * n = size of x, y, and alpha arrays
 * mod = modulus for reducing results
 */
static uint32_t 
interpolate(int i, int x[], uint32_t y[], unsigned n, uint32_t mod)
{
	u_int32_t	alpha[S4_KMAX];
	u_int32_t	prod;
	unsigned	l;
	unsigned	k;
	int		j;

	for (k = 0; k < n; k++) {
		alpha[k] = y[k];
	}
	for (k = 1; k < n; k++) {
		for (l = n - 1; l >= k; l--) {
			if (alpha[l] > alpha[l - 1]) {
				alpha[l] = alpha[l] - alpha[l - 1];
			} else {
				alpha[l] = alpha[l] - alpha[l - 1] + mod;
			}
			alpha[l] *= inverse(x[l] - x[l - k]);
			alpha[l] = alpha[l] % mod;
		}
	}
	prod = alpha[n - 1];
	for (j = n - 2; j >= 0; j--) {
		if (i < x[j]) {
			prod *= i - x[j] + mod;
		} else {
			prod *= i - x[j];
		}
		prod += alpha[j];
		prod %= mod;
	}
	return prod;
}

/*
 * Return a 16-bit-value from the file, with a flag set if only the
 * high 8 bits of the reconstructed value will be valid.  This is known
 * because odd-size files are padded with an extra byte.  We have to stay
 * two bytes ahead to know this.
 */
static uint32_t 
get_assemble_16(s4_t *s4, thresh_str_t *inputs, unsigned i, int *podd)
{
	uint32_t d1;
	int	c1;
	int	c2;

	if (!s4->seen[i]) {
		/* Get ahead the first time */
		c1 = inputs[i].io.base[inputs[i].c++];
		c2 = inputs[i].io.base[inputs[i].c++];
		s4->next_d1[i] = ((c1 & 0xff) << 8) + (c2 & 0xff);
		s4->seen[i] = 1;
	}
	d1 = s4->next_d1[i];
	if (inputs[i].c == inputs[i].io.size + 2) {
		s4->next_d1[i] = (unsigned)-1;
	} else {
		c1 = inputs[i].io.base[inputs[i].c++];
		if (inputs[i].c == inputs[i].io.size + 2) {
			*podd = 1;
			s4->next_d1[i] = (unsigned)-1;
		} else {
			c2 = inputs[i].io.base[inputs[i].c++];
			s4->next_d1[i] = ((c1 & 0xff) << 8) + (c2 & 0xff);
		}
	}
	return d1;
}

/*
 * Get 16-bit values from each file, and interpolate them to get the value
 * at x=0.  That is how we do the assembly.  Set the podd flag if
 * only the high 8 bits of this interpolated value are valid.
 */
static uint32_t
get_assemble(s4_t *s4, unsigned nin, thresh_str_t *inputs, int x[], uint32_t mod, int *podd)
{
	unsigned	i;
	int		y[S4_KMAX];

	for (i = 0; i < nin; i++) {
		if ((y[i] = (int)get_assemble_16(s4, inputs, i, podd)) == -1) {
			/* EOF */
			return (uint32_t)-1;
		}
	}
	return interpolate(0, x, (unsigned *)y, nin, mod);
}

/*
 * Given a set of nin file descriptors, assemble them to generate the
 * original file.  This is the main routine for the assembly case.
 */
static int
assemble(s4_t *s4, thresh_str_t *inputs, unsigned nin, uint32_t mod, thresh_str_t *output)
{
	thresh_head_t	heads[THRESH_MAX_SHARES];
	uint32_t	magic;
	uint32_t	c1;
	uint32_t	c2;
	unsigned	i;
	int		x[S4_KMAX];
	int		oddflag = 0;
	int		d;

	magic = S4_IMAGIC;
	for (i = 0; i < nin; i++) {
		thresh_get_header(&heads[i], inputs[i].io.base);
		if (!thresh_sane_header(&heads[i], &heads[0])) {
			(void) fprintf(stderr, "insane ssss header %d\n", i);
			return 0;
		}
		inputs[i].c += sizeof(heads[i]);
		x[i] = heads[i].coeff;
		//inputs[i].io.size += 2;
	}
	for (;;) {
		if ((d = get_assemble(s4, nin, inputs, x, mod, &oddflag)) == -1) {
			break;
		}
		if (d == (int)(mod - 1)) {
			d = get_assemble(s4, nin, inputs, x, mod, &oddflag);
			d += mod - 1;
		}
		d ^= magic;
		magic = (magic + S4_DMAGIC) & 0xffff;
		c1 = ((unsigned)d >> 8) & 0xff;
		output->io.base[output->c++] = c1;
		if (oddflag) {
			break;
		}
		c2 = d & 0xff;
		output->io.base[output->c++] = c2;
	}
	return 1;
}

/* check shares counts and threshold values are sane */
static int
check_values(unsigned threshold, unsigned sharesc)
{
	if (sharesc <= 1 || sharesc > S4_NMAX) {
		(void) fprintf(stderr, "Invalid number of shares (%d).\n", sharesc);
		return 0;
	}
	if (threshold < 1 || threshold > S4_NMAX) {
		(void) fprintf(stderr, "Invalid threshold number %d\n", threshold);
		return 0;
	}
	if (threshold > sharesc) {
		(void) fprintf(stderr, "threshold (%d) is larger than sharesc (%d)\n", threshold, sharesc);
		return 0;
	}
	return 1;
}

/* split the secret file into shares */
static int 
split_memory(threshold_t *thresh, const void *secret, unsigned memsize, unsigned threshold, unsigned sharesc)
{
	thresh_str_t	mem;
	unsigned	i;
	char		small[8];
	s4_t		*s4;

	s4 = thresh->handle;
	if (!check_values(threshold, sharesc)) {
		return 0;
	}
	(void) memset(&mem, 0x0, sizeof(mem));
	if ((mem.io.size = memsize) < S4_SMALLEST_INPUT) {
		(void) memset(small, 0x0, sizeof(small));
		(void) memcpy(small, secret, memsize = mem.io.size = S4_SMALLEST_INPUT);
		mem.io.base = (uint8_t *)small;
	} else {
		mem.io.base = (uint8_t *)__UNCONST(secret);
	}
	/* allocate space for outputs */
	thresh->sharesc = (unsigned)sharesc;
	for (i = 0 ; i < sharesc ; i++) {
		thresh->shares[i].io.size = memsize + sizeof(thresh_head_t);
		thresh->shares[i].c = 0;
		if (thresh->shares[i].io.base == NULL &&
		    (thresh->shares[i].io.base = calloc(1, thresh->shares[i].io.size)) == NULL) {
			(void) fprintf(stderr, "Bad alloc\n");
			return 0;
		}
	}
	/* Do the work */
	split(s4, &mem, sharesc, threshold, PRIME, thresh->shares);
	return 1;
}

/* reconstruct the shares from memory */
static int 
join_memory(threshold_t *thresh, thresh_str_t *input, unsigned bytes, unsigned filec, thresh_str_t *str)
{
	s4_t	*s4;

	s4 = thresh->handle;
	(void) memset(str, 0x0, sizeof(*str));
	if (str->io.base == NULL &&
	    (str->io.base = calloc(1, str->io.size = bytes - sizeof(thresh_head_t))) == NULL) {
		(void) fprintf(stderr, "Bad alloc %u bytes\n", bytes);
		return 0;
	}
	/* Do the work */
	return assemble(s4, input, filec, PRIME, str);
}

/******************** exported functions from here on ********************/

/* split the memory */
int
s4_split(threshold_t *thresh)
{
	thresh_str_t	*mem;

	mem = &thresh->shares[THRESH_MAX_SHARES];
	return split_memory(thresh, mem->io.base, mem->io.size, thresh->threshold, thresh->sharesc);
}

/* single combine function */
int
s4_combine(threshold_t *thresh)
{
	if (!join_memory(thresh, thresh->shares, 
			thresh->shares[0].io.size, thresh->threshold,
			&thresh->shares[THRESH_JOIN_DEST])) {
		return -1;
	}
	return thresh->shares[THRESH_MAX_SHARES].c;
}

/* initialise for s4 */
int
s4_init(threshold_t *threshold, unsigned thresh, unsigned sharesc)
{
	s4_t	*s4;

	if ((s4 = calloc(1, sizeof(*s4))) == NULL) {
		fprintf(stderr, "Unable to allocate ssss structure\n");
		return 0;
	}
	threshold->handle = s4;
	s4->magic = S4_IMAGIC;
	s4->sharesc = S4_KMAX; /* guess for just now */
	return 1;
}

/* free up resources */
int
s4_end(threshold_t *thresh)
{
	s4_t		*s4;

	s4 = thresh->handle;
	free(s4);
	thresh->handle = NULL;
	return 1;
}
